#include <stdlib.h>
#include <stdio.h>
#include <sys/times.h>
#include <sys/types.h>
#include <pthread.h>

typedef struct {
	void *(*routine)(void*);
	void *arg;
}dispatcherArg;

pthread_mutex_t *fileMutex;
FILE *file;

void *dispatcher(void *arg) {
	struct tms initialTimeMarker, finalTimeMarker;
	clock_t initialUtime, initialStime, finalUtime, finalStime;
	dispatcherArg *dArg;
	pthread_t threadId;
	
	dArg = (dispatcherArg *)arg;
	
	times(&initialTimeMarker);
	initialUtime = initialTimeMarker.tms_utime;
	initialStime = initialTimeMarker.tms_stime;
	
	(dArg->routine)(dArg->arg);
	
	times(&finalTimeMarker);
	finalUtime = finalTimeMarker.tms_utime;
	finalStime = finalTimeMarker.tms_stime;

	threadId = pthread_self();
	
	pthread_mutex_lock(fileMutex);
	
	fprintf(file, "1;%lu;%ld;%ld;%ld;%ld\n", threadId, initialUtime, initialStime, finalUtime, finalStime);
	
	pthread_mutex_unlock(fileMutex);
	
	free(dArg);
	
	return 0;
}

int thread_create(pthread_t *thread, pthread_attr_t *attr, void *(*start_routine)(void*), void *arg) {
	int result;
	
	dispatcherArg *dArg;
	dArg = (dispatcherArg*)malloc(sizeof(dispatcherArg));
	dArg->routine = start_routine;
	dArg->arg = arg;
	
	result = pthread_create(thread, attr, dispatcher, dArg);
	
	return result;
}

int mutex_lock(pthread_mutex_t * mutex) {
	pthread_t threadId;
	struct tms timeMarker;
	clock_t initialUtime, initialStime, finalUtime, finalStime;
	int result;
	
	times(&timeMarker);
	initialUtime = timeMarker.tms_utime;
	initialStime = timeMarker.tms_stime;
	threadId = pthread_self();
	
	result = pthread_mutex_lock(mutex);
	
	times(&timeMarker);
	finalUtime = timeMarker.tms_utime;
	finalStime = timeMarker.tms_stime;
	
	pthread_mutex_lock(fileMutex);
	
	fprintf(file, "2;%lu;%lu;%ld;%ld;%ld;%ld\n", threadId, mutex, initialUtime, initialStime, finalUtime, finalStime);
	
	pthread_mutex_unlock(fileMutex);
	
	return result;
}

int mutex_unlock(pthread_mutex_t * mutex) {
	pthread_t threadId;
	struct tms timeMarker;
	clock_t utime, stime;
	
	times(&timeMarker);
	utime = timeMarker.tms_utime;
	stime = timeMarker.tms_stime;
	threadId = pthread_self();
	
	pthread_mutex_lock(fileMutex);
	
	fprintf(file, "3;%lu;%lu;%ld;%ld\n", threadId, mutex, utime, stime);
	
	pthread_mutex_unlock(fileMutex);
	
	return pthread_mutex_unlock(mutex);
}

int thread_join(pthread_t thread, void **value_ptr) {
	pthread_t callerThreadId;
	struct tms timeMarker;
	clock_t utime, stime;
	
	times(&timeMarker);
	utime = timeMarker.tms_utime;
	stime = timeMarker.tms_stime;
	callerThreadId = pthread_self();
	
	pthread_mutex_lock(fileMutex);
	
	fprintf(file, "4;%lu;%lu;%ld;%ld\n", callerThreadId, thread, utime, stime);
	
	pthread_mutex_unlock(fileMutex);
	
	return pthread_join(thread, value_ptr);
}

void initializeVariables(char *fileName) {
	if(fileMutex == NULL) {		
		fileMutex = (pthread_mutex_t*)malloc(sizeof(pthread_mutex_t));
		
		pthread_mutex_init(fileMutex, NULL);
		file = fopen(fileName, "w");
	}
}

void destroyVariables() {
	if(fileMutex != NULL) {		
		pthread_mutex_destroy(fileMutex);
		free(fileMutex);
		fclose(file);
	}
}
